#include <dirent.h>
#include <fstream>
#include <pwd.h>
#include <signal.h>
#include <string>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <vector>

static vector<int> waitpids;
static unsigned slplen = 1;
static class env
{
    const char *q;
public:
    env() : q(NULL) { }
    env(const char *p) : q(p) { }
    env operator[] (const char *s) const { return env(s); }
    void operator= (const char *v) { setenv(q, v, 1); }
    operator const char *() { return getenv(q); }
} env;

class envforcer
{
public:
    string var, val;
    envforcer(const string &r, const string &l) : var(r), val(l)
    {
    }
};
static vector<envforcer> envforce;

class entry
{
public:
    vector<bool> mins, hours, day, mon, wday;
    string command;
    
    void set(vector<bool> &dest, const string &s);
    
    entry(const string &s) : mins(60, false),
                             hours(24, false),
                             day(32, false),
                             mon(13, false),
                             wday(7, false)
    {
        string t;
        unsigned a=0, b=s.size(), c;
        for(c=0; c<5; ++c)
        {
            for(t=""; a<b && s[a]!=' ' && s[a]!='\t'; ++a)t += s[a];
            while(a<b && s[a]==' ' || s[a]=='\t')a++;
            switch(c)
            {
                case 0: set(mins, t); break;
                case 1: set(hours, t); break;
                case 2: set(day, t); break;
                case 3: set(mon, t); break;
                case 4: set(wday, t); break;
            }
        }
        command = s.substr(a);
    }
    
    void action(const struct passwd *pwd)
    {
        int p = fork();
        if(p)
        {
            if(p > 0)
            {
                waitpids.push_back(p);
                slplen = 1;
            }
            return;
        }
        
        setuid(pwd->pw_uid);
        setgid(pwd->pw_gid);
        
        env["SHELL"]   = "/bin/sh";
        env["HOME"]    = pwd->pw_dir;
        env["LOGNAME"] = pwd->pw_name;
        
        for(unsigned a=0; a<envforce.size(); ++a)
            env[envforce[a].var.c_str()] = envforce[a].val.c_str();
        
        if(execl(env["SHELL"], "sh", "-c", command.c_str(), NULL) < 0)
        {
            cerr << "bcrond: Can not execute " << command << endl;
        }
        _exit(0);
    }
    
    void test(const struct tm *const TM, const struct passwd *p)
    {
        if(!mins[TM->tm_min])return;
        if(!hours[TM->tm_hour])return;
        if(wday[TM->tm_wday]
        || (TM->tm_wday == 0 && wday[7])
        || (mon[TM->tm_mon+1] && day[TM->tm_mday]))
            action(p);
    }
};

void entry::set(vector<bool> &dest, const string &s)
{
    unsigned a=0;
    if(s=="*")
        for(; a<dest.size(); ++a)dest[a]=true;
    else
    {
        unsigned n, b=s.size(), c=0;
        bool pilkku = true;
        for(; a<b; ++a)
        {
            for(n=0; a<b && s[a]>='0' && s[a]<='9'; ++a)
                n = n*10 + (s[a] - '0');
                
            if(n < dest.size())
            {
                if(!pilkku)while(c<n)dest[c++] = true;
                dest[c=n] = true;
            }
                   
            if(s[a] == '-')
                pilkku = false;
            else
            {
                if(s[a] != ',')break;
                pilkku = true;
            }
        }
    }
}


static void waiters()
{
    for(;;)
    {
        int p = waitpid(-1, NULL, WNOHANG);
        if(p <= 0)break;
        
        vector<int>::iterator i;
        for(i=waitpids.begin(); i!=waitpids.end(); ++i)
            if(*i == p)
            {
                waitpids.erase(i);
                break;
            }
    }
}

static void handlefile(const struct tm *const TM, const char *ss, struct passwd *pwd = NULL)
{
	if(!pwd)
	{
		pwd = getpwuid(0);
		if(!pwd)return; /* Paranoid */
	}
    ifstream f(ss);
    envforce.clear();
    for(;;)
    {
        string s;
        getline(f, s);
        if(!f.good())break;
        if(s.substr(0, 1) == "#" || !s.size())continue;
        unsigned equpos = s.find('=');
        unsigned spcpos = s.find(' ');
        if(equpos != s.npos && equpos < spcpos)
        {
            envforcer tmp(s.substr(0, equpos), s.substr(equpos+1));
            envforce.push_back(tmp);
            continue;
        }
        entry diu(s);
        diu.test(TM, pwd);
    }
}

static void tick(const struct tm *const TM)
{
    handlefile(TM, "/etc/crontab");
    DIR *dir = opendir(DATADIR);
    if(!dir)return;
    struct dirent *ent;
    while((ent = readdir(dir)))
    {
        struct passwd *pwd = getpwnam(ent->d_name);
        if(!pwd)
        {
        	// Skip entries which are of no users
        	// This includes backup~ files and log
        	continue;
        }        
        handlefile(TM, (string(DATADIR"/") + ent->d_name).c_str(), pwd);
    }
    closedir(dir);
}

static void help()
{    
    cout <<
        "Bisqwit's crond version "VERSION"\n"
        "Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n"
        "\n"
        "Usage: bcrond [-f]\n"
        "    -f = Don't fork (stay foreground)\n"
        "\n"
        "Uses the data in "DATADIR" directory.\n";
}

int main(int argc, const char *const *argv)
{
    bool background=true;
    bool error     =false;
    
    while(--argc)
    {
        const char *s = *++argv;
        if(*s == '-')
            while(*++s)
                switch(*s)
                {
                    case 'f': background=false; break;
                    default: error=true;
                }
        else
            error=true;
    }
    if(error)
    {
        help();
        return 0;
    }
    if(background)
    {
        chdir("/");
        // This is correct, believe or not.
        if(fork() || (setsid() >= 0 && fork()))exit(0);
    }

    // Simply ignore it.
    signal(SIGHUP, SIG_IGN);
    
    int min = -1;
    for(;;)
    {
        time_t t = time(NULL);
        struct tm *TM = localtime(&t);
        if(TM->tm_min != min)
        {
            min = TM->tm_min;
            tick(TM);
            if(!waitpids.size())
                slplen = 30;
        }
        sleep(slplen);
        waiters();
    }
}
